@@ -23,7 +23,7 @@ use GRID::Machine::MakeAccessors; # Order is important. This must be the first!
 use GRID::Machine::Message;
 use GRID::Machine::Result;
 
-our $VERSION = "0.081";
+our $VERSION = "0.082";
 
 sub read_modules {
 
@@ -593,6 +593,27 @@ sub _get_result_or_callback {
 }
 
 # True if machine accepts automatic ssh connections 
+#sub is_operative {
+#  my $ssh = shift;
+#  my $host = shift;
+#  my $command = shift || 'perl -v';
+#  my $seconds = shift || 10;
+#
+#    my $operative;
+#
+#    my $devnull = File::Spec->devnull();
+#    local $SIG{ALRM} = sub { $operative = 0 };
+#    alarm($seconds);
+#    eval {
+#      # Warning O.S. dependent code!
+#      $operative = !system("$ssh $host $command > $devnull < $devnull");
+#    };
+#    alarm(0);
+#    return $operative;
+#}
+
+# True if machine accepts automatic ssh connections 
+# Eric version. Thanks Eric!
 sub is_operative {
   my $ssh = shift;
   my $host = shift;
@@ -602,13 +623,17 @@ sub is_operative {
     my $operative;
 
     my $devnull = File::Spec->devnull();
-    local $SIG{ALRM} = sub { $operative = 0 };
-    alarm($seconds);
     eval {
-      # Warning O.S. dependent code!
-      $operative = !system("$ssh $host $command > $devnull < $devnull");
+        local $SIG{ALRM} = sub { die "Alarm exceeded $seconds seconds"; };
+        alarm($seconds);
+        # Warning O.S. dependent code!
+        $operative = !system("$ssh $host $command > $devnull < $devnull");
+        alarm(0);
     };
-    alarm(0);
+
+    if($@) {
+        return 0;
+    }
     return $operative;
 }
 
